/*  Copyright 2003-2013 Ciaran Anscomb
 *
 *  This file is part of XRoar.
 *
 *  XRoar is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  XRoar is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with XRoar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

// For strsep()
#define _BSD_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL.h>

#include "pl_glib.h"
#include "pl_string.h"

#include "joystick.h"
#include "keyboard.h"
#include "logging.h"
#include "machine.h"
#include "module.h"
#include "printer.h"
#include "xroar.h"

#include "sdl/common.h"

static _Bool init(void);
static void update_kbd_translate(void);

KeyboardModule keyboard_sdl_module = {
	.common = { .name = "sdl", .description = "SDL keyboard input",
	            .init = init },
	.update_kbd_translate = update_kbd_translate,
};

struct keymap {
	const char *name;
	unsigned int *raw;
};

#include "keyboard_sdl_mappings.c"

static unsigned int control = 0, shift = 0;
static _Bool noratelimit_latch = 0;

static uint_least16_t sdl_to_keymap[768];

/* Track which unicode value was last generated by key presses (SDL only
 * guarantees to fill in the unicode field on key presses, not releases). */
static unsigned int unicode_last_keysym[SDLK_LAST];

static const char *keymap_option;
static unsigned int *selected_keymap;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static struct joystick_axis *configure_axis(char *, unsigned);
static struct joystick_button *configure_button(char *, unsigned);
static void unmap_axis(struct joystick_axis *axis);
static void unmap_button(struct joystick_button *button);

struct joystick_interface sdl_js_if_keyboard = {
	.name = "keyboard",
	.configure_axis = configure_axis,
	.configure_button = configure_button,
	.unmap_axis = unmap_axis,
	.unmap_button = unmap_button,
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

struct axis {
	SDLKey key0, key1;
	unsigned value;
};

struct button {
	SDLKey key;
	_Bool value;
};

#define MAX_AXES (4)
#define MAX_BUTTONS (4)

static struct axis *enabled_axis[MAX_AXES];
static struct button *enabled_button[MAX_BUTTONS];

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static void map_keyboard(unsigned int *map) {
	int i;
	for (i = 0; i < 256; i++)
		sdl_to_keymap[i] = i & 0x7f;
	for (i = 0; i < SDLK_LAST; i++)
		unicode_last_keysym[i] = 0;
	if (map == NULL)
		return;
	while (*map) {
		unsigned int sdlkey = *(map++);
		unsigned int dgnkey = *(map++);
		sdl_to_keymap[sdlkey & 0xff] = dgnkey & 0x7f;
	}
}

static _Bool init(void) {
	keymap_option = xroar_cfg.keymap;
	selected_keymap = NULL;
	for (unsigned i = 0; mappings[i].name; i++) {
		if (selected_keymap == NULL
				&& !strcmp("uk", mappings[i].name)) {
			selected_keymap = mappings[i].raw;
		}
		if (keymap_option && !strcmp(keymap_option, mappings[i].name)) {
			selected_keymap = mappings[i].raw;
			LOG_DEBUG(2,"\tSelecting '%s' keymap\n",keymap_option);
		}
	}
	map_keyboard(selected_keymap);
	SDL_EnableUNICODE(xroar_cfg.kbd_translate);

	for (unsigned i = 0; i < MAX_AXES; i++)
		enabled_axis[i] = NULL;
	for (unsigned i = 0; i < MAX_BUTTONS; i++)
		enabled_button[i] = NULL;

	return 1;
}

static void update_kbd_translate(void) {
	SDL_EnableUNICODE(xroar_cfg.kbd_translate);
}

static void emulator_command(SDLKey sym) {
	switch (sym) {
	case SDLK_1: case SDLK_2: case SDLK_3: case SDLK_4:
		if (shift) {
			xroar_new_disk(sym - SDLK_1);
		} else {
			xroar_insert_disk(sym - SDLK_1);
		}
		break;
	case SDLK_5: case SDLK_6: case SDLK_7: case SDLK_8:
		if (shift) {
			xroar_select_write_back(sym - SDLK_5, XROAR_TOGGLE);
		} else {
			xroar_select_write_enable(sym - SDLK_5, XROAR_TOGGLE);
		}
		break;
	case SDLK_a:
		xroar_select_cross_colour(XROAR_CYCLE);
		break;
	case SDLK_c:
	case SDLK_q:
		xroar_quit();
		break;
	case SDLK_e:
		xroar_toggle_cart();
		break;
	case SDLK_f:
		xroar_fullscreen(XROAR_TOGGLE);
		break;
	case SDLK_i:
		if (shift) {
			xroar_load_file(xroar_cart_exts);
		} else {
			xroar_run_file(xroar_cart_exts);
		}
		break;
	case SDLK_j:
		if (shift) {
			joystick_swap();
		} else {
			joystick_cycle();
		}
		break;
	case SDLK_k:
		xroar_set_keymap(XROAR_CYCLE);
		break;
	case SDLK_b:
	case SDLK_h:
	case SDLK_t:
	case SDLK_l:
		if (shift) {
			xroar_run_file(NULL);
		} else {
			xroar_load_file(NULL);
		}
		break;
	case SDLK_m:
		xroar_set_machine(XROAR_CYCLE);
		break;
	case SDLK_p:
		if (shift)
			printer_flush();
		break;
	case SDLK_r:
		if (shift)
			xroar_hard_reset();
		else
			xroar_soft_reset();
		break;
	case SDLK_s:
		xroar_save_snapshot();
		break;
	case SDLK_w:
		xroar_select_tape_output();
		break;
#ifdef TRACE
	case SDLK_v:
		xroar_set_trace(XROAR_TOGGLE);
		break;
#endif
	case SDLK_z: /* running out of letters... */
		xroar_set_kbd_translate(XROAR_TOGGLE);
		break;
	default:
		break;
	}
	return;
}

void sdl_keypress(SDL_keysym *keysym) {
	SDLKey sym = keysym->sym;

	for (unsigned i = 0; i < MAX_AXES; i++) {
		if (enabled_axis[i]) {
			if (sym == enabled_axis[i]->key0) {
				enabled_axis[i]->value = 0;
				return;
			}
			if (sym == enabled_axis[i]->key1) {
				enabled_axis[i]->value = 255;
				return;
			}
		}
	}
	for (unsigned i = 0; i < MAX_BUTTONS; i++) {
		if (enabled_button[i]) {
			if (sym == enabled_button[i]->key) {
				enabled_button[i]->value = 1;
				return;
			}
		}
	}

	if (sym == SDLK_LSHIFT || sym == SDLK_RSHIFT) {
		shift = 1;
		KEYBOARD_PRESS_SHIFT;
		return;
	}
	if (sym == SDLK_LCTRL || sym == SDLK_RCTRL) { control = 1; return; }
	if (sym == SDLK_F11) {
		xroar_fullscreen(XROAR_TOGGLE);
		return;
	}
	if (sym == SDLK_F12) {
		if (shift) {
			noratelimit_latch = !noratelimit_latch;
			if (noratelimit_latch) {
				xroar_noratelimit = 1;
				xroar_frameskip = 10;
			} else {
				xroar_noratelimit = 0;
				xroar_frameskip = xroar_cfg.frameskip;
			}
		} else if (!noratelimit_latch) {
			xroar_noratelimit = 1;
			xroar_frameskip = 10;
		}
	}
	if (sym == SDLK_PAUSE) { machine_toggle_pause(); return; }
	if (control) {
		emulator_command(sym);
		return;
	}
	if (sym == SDLK_UP) { KEYBOARD_PRESS(KEYMAP_UP); return; }
	if (sym == SDLK_DOWN) { KEYBOARD_PRESS(KEYMAP_DOWN); return; }
	if (sym == SDLK_LEFT) { KEYBOARD_PRESS(KEYMAP_LEFT); return; }
	if (sym == SDLK_RIGHT) { KEYBOARD_PRESS(KEYMAP_RIGHT); return; }
	if (sym == SDLK_HOME) { KEYBOARD_PRESS_CLEAR; return; }
	if (xroar_cfg.kbd_translate) {
		unsigned int unicode;
		if (sym >= SDLK_LAST)
			return;
		unicode = keysym->unicode;
		/* Map shift + backspace/delete to ^U */
		if (shift && (unicode == 8 || unicode == 127)) {
			unicode = 0x15;
		}
		unicode_last_keysym[sym] = unicode;
		keyboard_unicode_press(unicode);
		return;
	}
	if (sym == SDLK_WORLD_18) sym = KEYMAP_CLEAR;
	if (sym < 256) {
		unsigned int mapped = sdl_to_keymap[sym];
		KEYBOARD_PRESS(mapped);
	}
}

void sdl_keyrelease(SDL_keysym *keysym) {
	SDLKey sym = keysym->sym;

	for (unsigned i = 0; i < MAX_AXES; i++) {
		if (enabled_axis[i]) {
			if (sym == enabled_axis[i]->key0) {
				if (enabled_axis[i]->value < 129)
					enabled_axis[i]->value = 129;
				return;
			}
			if (sym == enabled_axis[i]->key1) {
				if (enabled_axis[i]->value > 130)
					enabled_axis[i]->value = 130;
				return;
			}
		}
	}
	for (unsigned i = 0; i < MAX_BUTTONS; i++) {
		if (enabled_button[i]) {
			if (sym == enabled_button[i]->key) {
				enabled_button[i]->value = 0;
				return;
			}
		}
	}

	if (sym == SDLK_LSHIFT || sym == SDLK_RSHIFT) {
		shift = 0;
		KEYBOARD_RELEASE_SHIFT;
		return;
	}
	if (sym == SDLK_LCTRL || sym == SDLK_RCTRL) { control = 0; return; }
	if (sym == SDLK_F12) {
		if (!noratelimit_latch) {
			xroar_noratelimit = 0;
			xroar_frameskip = xroar_cfg.frameskip;
		}
	}
	if (sym == SDLK_UP) { KEYBOARD_RELEASE(KEYMAP_UP); return; }
	if (sym == SDLK_DOWN) { KEYBOARD_RELEASE(KEYMAP_DOWN); return; }
	if (sym == SDLK_LEFT) { KEYBOARD_RELEASE(KEYMAP_LEFT); return; }
	if (sym == SDLK_RIGHT) { KEYBOARD_RELEASE(KEYMAP_RIGHT); return; }
	if (sym == SDLK_HOME) { KEYBOARD_RELEASE_CLEAR; return; }
	if (xroar_cfg.kbd_translate) {
		unsigned int unicode;
		if (sym >= SDLK_LAST)
			return;
		unicode = unicode_last_keysym[sym];
		/* Map shift + backspace/delete to ^U */
		if (shift && (unicode == 8 || unicode == 127)) {
			unicode = 0x15;
		}
		keyboard_unicode_release(unicode);
		/* Might have unpressed shift prematurely */
		if (shift)
			KEYBOARD_PRESS_SHIFT;
		return;
	}
	if (sym == SDLK_WORLD_18) sym = KEYMAP_CLEAR;
	if (sym < 256) {
		unsigned int mapped = sdl_to_keymap[sym];
		KEYBOARD_RELEASE(mapped);
	}
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static unsigned read_axis(struct axis *a) {
	return a->value;
}

static _Bool read_button(struct button *b) {
	return b->value;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static SDLKey get_key_by_name(const char *name) {
	if (isdigit(name[0]))
		return strtol(name, NULL, 0);
	for (SDLKey i = SDLK_FIRST; i < SDLK_LAST; i++) {
		if (0 == g_ascii_strcasecmp(name, SDL_GetKeyName(i)))
			return i;
	}
	return SDLK_UNKNOWN;
}

static struct joystick_axis *configure_axis(char *spec, unsigned jaxis) {
	SDLKey key0, key1;
	// sensible defaults
	if (jaxis == 0) {
		key0 = SDLK_LEFT;
		key1 = SDLK_RIGHT;
	} else {
		key0 = SDLK_UP;
		key1 = SDLK_DOWN;
	}
	char *a0 = NULL;
	char *a1 = NULL;
	if (spec) {
		a0 = strsep(&spec, ",");
		a1 = spec;
	}
	if (a0 && *a0)
		key0 = get_key_by_name(a0);
	if (a1 && *a1)
		key1 = get_key_by_name(a1);
	struct axis *axis_data = g_malloc(sizeof(*axis_data));
	axis_data->key0 = key0;
	axis_data->key1 = key1;
	axis_data->value = 127;
	struct joystick_axis *axis = g_malloc(sizeof(*axis));
	axis->read = (js_read_axis_func)read_axis;
	axis->data = axis_data;
	for (unsigned i = 0; i < MAX_AXES; i++) {
		if (!enabled_axis[i]) {
			enabled_axis[i] = axis_data;
			break;
		}
	}
	return axis;
}

static struct joystick_button *configure_button(char *spec, unsigned jbutton) {
	SDLKey key = (jbutton == 0) ? SDLK_LALT : SDLK_UNKNOWN;
	if (spec && *spec)
		key = get_key_by_name(spec);
	struct button *button_data = g_malloc(sizeof(*button_data));
	button_data->key = key;
	button_data->value = 0;
	struct joystick_button *button = g_malloc(sizeof(*button));
	button->read = (js_read_button_func)read_button;
	button->data = button_data;
	for (unsigned i = 0; i < MAX_BUTTONS; i++) {
		if (!enabled_button[i]) {
			enabled_button[i] = button_data;
			break;
		}
	}
	return button;
}

static void unmap_axis(struct joystick_axis *axis) {
	if (!axis)
		return;
	for (unsigned i = 0; i < MAX_AXES; i++) {
		if (axis->data == enabled_axis[i]) {
			enabled_axis[i] = NULL;
		}
	}
	g_free(axis->data);
	g_free(axis);
}

static void unmap_button(struct joystick_button *button) {
	if (!button)
		return;
	for (unsigned i = 0; i < MAX_BUTTONS; i++) {
		if (button->data == enabled_button[i]) {
			enabled_button[i] = NULL;
		}
	}
	g_free(button->data);
	g_free(button);
}
